/*
 * Handles hardware mapped registers access.
 * 
 * Approximately the same API as the memory access module.
 *
 * Copyright (c) 2009-2010 Pierre "delroth" Bourdon <delroth@gmail.com>
 *
 * This file is part of Meta-morphosis.
 *
 * Meta-morphosis is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Meta-morphosis is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Meta-morphosis.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "gpu.h"
#include "hardware.h"
#include "memory.h"
#include "msg.h"
#include "spu.h"

int log_unhandled_hw = 0;

#define FAIL(act) \
    if (log_unhandled_hw) \
    { \
        psx_message("unable to " #act " hardware at 0x%04X\n", addr); \
    } \
    else \
        (void)0

#define RFAIL FAIL(read)
#define WFAIL FAIL(write)

uint8_t psx_hardware_read_8(uint16_t addr)
{
    RFAIL;

    return psx_mem_read_8_direct(0x1F80000 + addr);
}

uint16_t psx_hardware_read_16(uint16_t addr)
{
    if (addr >= 0x1c00 && addr < 0x1e00)
    {
        return psx_spu_read_register(0x1F800000 & addr);
    }

    RFAIL;

    return psx_mem_read_16_direct(0x1F800000 + addr);
}

uint32_t psx_hardware_read_32(uint16_t addr)
{
    if (addr == 0x1810) return psx_gpu_read_data();
    if (addr == 0x1814) return psx_gpu_read_status();

    RFAIL;

    return psx_mem_read_32_direct(0x1F800000 + addr);
}

#define AT(a, act) \
    if (addr == a) { act; return; } else (void)0

void psx_hardware_write_8(uint16_t addr, uint8_t data)
{
    WFAIL;
}

void psx_hardware_write_16(uint16_t addr, uint16_t data)
{
    if (addr >= 0x1c00 && addr < 0x1e00)
    {
        psx_spu_write_register(0x1F800000 & addr, data);
        return;
    }

    WFAIL;
}

typedef void (*dma_handler)(uint32_t, uint32_t, uint32_t);

static inline void do_dma(dma_handler h, uint8_t n, uint32_t val, 
                          uint32_t madr, uint32_t bcr, uint32_t chcr)
{
    uint32_t vchcr = psx_mem_read_32_direct(chcr);
    uint32_t pcr = psx_mem_read_32_direct(0x1F8010F0);
    madr = psx_mem_read_32_direct(madr);
    bcr = psx_mem_read_32_direct(bcr);

    if (vchcr & 0x01000000) /* DMA already occuring */
        return;

    psx_mem_write_32_direct(chcr, val);

    if ((val & 0x01000000) && (pcr & (8 << (n * 4))))
        h(madr, bcr, val);

    psx_mem_write_32_direct(chcr, ~0x01000000);
}

void psx_hardware_write_32(uint16_t addr, uint32_t data)
{
    AT(0x10a8, do_dma(psx_gpu_do_dma, 2, data,
                      0x1F8010A0, 0x1F8010A4, 0x1F8010A8));
    AT(0x1810, psx_gpu_write_data(data));
    AT(0x1814, psx_gpu_write_status(data));


    WFAIL;

    psx_mem_write_32_direct(0x1F800000 + addr, data);
}
