/*
 * Defines the functions used to emulate the R3000 instructions.
 *
 * Copyright (c) 2009-2010 Pierre "delroth" Bourdon <delroth@gmail.com>
 *
 * This file is part of Meta-morphosis.
 *
 * Meta-morphosis is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Meta-morphosis is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Meta-morphosis.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "address.h"
#include "../bios/bios.h"
#include "../bios/exception.h"
#include "codegate.h"
#include "globals.h"
#include "gpu.h"
#include "hardware.h"
#include "interpreter.h"
#include "memory.h"

#include <stdint.h>
#include <sys/time.h>
#include <unistd.h>

/*
 * Get / set registers.
 */
uint32_t psx_get_reg(int n)
{
    return GPR[n];
}

/* VSync time handling */

#define VSYNC_INTERVAL (1000000 / 50) /* 60 FPS */

/* Command line option */
int count_vsync = 0;

static double vsync_mean = 0.0;
static uint32_t vsync_wait_count = 0;
static uint64_t vsync_count = 0;

static struct timeval vsync_tv = { 0, 0 };

void vsync_time_reset(void)
{
    static struct timezone tz;

    if (count_vsync && vsync_count)
    {
        vsync_mean = (vsync_mean * vsync_count + vsync_wait_count);
        vsync_mean /= ++vsync_count;
    }
    else
    {
        vsync_mean = vsync_wait_count;
    }

    vsync_wait_count = 0;
    gettimeofday(&vsync_tv, &tz);
}

int vsync_time_elapsed(void)
{
    static struct timezone tz;
    struct timeval tv;

    uint64_t totusec1, totusec2;

    gettimeofday(&tv, &tz);

    totusec1 = tv.tv_sec * 1000000 + tv.tv_usec;
    totusec2 = vsync_tv.tv_sec * 1000000 + vsync_tv.tv_usec;

    return (totusec1 > (totusec2 + VSYNC_INTERVAL));
}

double psx_get_vsync_wait_mean(void)
{
    return vsync_mean;
}

void psx_set_reg(int n, uint32_t v)
{
    GPR[n] = v;
}

/* Handle hardware interrupts */
static inline void psx_handle_interrupts(void)
{
    /* if (psx_hardware_read_32(0x1070) & psx_hardware_read_32(0x1074)) */
    if (vsync_time_elapsed())
    {
        psx_bios_exception();
        psx_gpu_update();
        vsync_time_reset();
    }
    else
        vsync_wait_count += 1;
}

/* Handles jumps - delay shots, etc. */
void psx_jump(uint32_t addr)
{
    /* TODO : Handle delays, etc. */
    psx_exec_one();

    pc = addr;

    psx_handle_interrupts();
}

static inline void jump_and_link(uint32_t addr)
{
    /* 31 = RA, return address, aka. link register */
    GPR[31] = pc + 4;
    psx_jump(addr);
}

/* Implementation of theinstructions */

#define INSTR(n) \
void psx_instr_##n(uint32_t a, uint32_t b, uint32_t c, uint32_t d)

/* Special instructions : call to BIOS, etc. */
void psx_instr_callbios(void)
{
    psx_bios_entry((pc >> 4) - 0xA, GPR);

    /* Jump back to the old PC */
    pc = GPR[RA];
}
