(*
 * Compiles a Psx_file.t to unoptimized LLVM bytecode.
 *
 * Basically, it translates every instruction to a function call.
 *
 * Copyright (c) 2009-2010 Pierre "delroth" Bourdon <delroth@gmail.com>
 *
 * This file is part of Meta-morphosis.
 *
 * Meta-morphosis is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Meta-morphosis is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Meta-morphosis.  If not, see <http://www.gnu.org/licenses/>.
 *)

#include "../config.h"

module D = R3000_decoder
module P = Psx_file
module Reg = R3000_reg

type state = {
    ctx : Llvm.llcontext ;
    m : Llvm.llmodule ;
    b : Llvm.llbuilder ;
    mutable pc : Llvm.llvalue ;
    f : Psx_file.t
}

(* LLVM is imperative, we'll use the module in the state to output code *)
type out = state

let init f =
    let ctx = Llvm.create_context () in
    let b = Llvm.builder ctx in
    let m = Llvm.create_module ctx "mmorph_main" in
    { ctx = ctx ; m = m ; f = f ; b = b ; pc = Obj.magic () }

let to_bytecode o chan =
    Llvmext.write_bitcode_to_chan o.m chan

let copy ic oc =
    let eof = ref false in
    let buf = String.create 1024 in
    while not !eof do
        let l = input ic buf 0 1024 in
        if l <> 1024 then
            eof := true ;
        output oc buf 0 l
    done

let to_binary o chan =
    let llc_out, llc_in = Unix.open_process (LLC_PATH ^ " -filetype=asm") in
    let out = Filename.temp_file "translator" ".exe" in
    let gcc_cmd = Printf.sprintf
        "%s -x %s -o %s -L%s -lpthread -lpsx -lbios -"
        CC_PATH "assembler" out LIBDIR in
    let gcc_out, gcc_in = Unix.open_process gcc_cmd in
    prerr_endline "Translating the IL into ASM..." ;
    to_bytecode o llc_in ;
    close_out llc_in ;
    prerr_endline "GCC is linking the output program..." ;
    copy llc_out gcc_in ;
    ignore (Unix.close_process (gcc_out, gcc_in)) ;
    copy (open_in_bin out) chan ;
    Unix.unlink out

let to_il o chan =
    Llvmext.print_module o.m chan

(* ---------------------------------------------------------------------- *)

let define_pc st =
    let it = Llvm.i32_type st.ctx in
    let pc = Int64.of_int32 st.f.P.header.P.init_pc in
    let v = Llvm.const_of_int64 it pc false in
    ignore (Llvm.define_global "pc" v st.m)

let define_segs st =
    let it = Llvm.i32_type st.ctx in
    let bt = Llvm.i8_type st.ctx in
    let def_i32 n v =
        let v = Int64.of_int32 v in
        ignore (Llvm.define_global n (Llvm.const_of_int64 it v false) st.m)
    in
    let def_seg n v =
        let l = Array.length v in
        let arr = Array.init (4 * l) (fun i ->
            let shift = [| 0 ; 8 ; 16 ; 24 |] in
            let x = Int32.shift_right_logical v.(i / 4) shift.(i mod 4) in
            let x = Int32.logand x 0xFFl in
            Llvm.const_int bt (Int32.to_int x)
        ) in
        let v = Llvm.const_array bt arr in
        ignore (Llvm.define_global n v st.m)
    in
    def_seg "t_seg" st.f.P.code ;
    def_i32 "t_seg_addr" st.f.P.header.P.text.P.addr ;
    def_i32 "t_seg_size" st.f.P.header.P.text.P.size ;
    def_i32 "s_seg_addr" st.f.P.header.P.stack.P.addr ;
    def_i32 "init_gp" st.f.P.header.P.init_gp

let define_gpr st =
    let it = Llvm.i32_type st.ctx in
    let vals = Array.init 32 (fun _ -> Llvm.const_int it 0) in
    let v = Llvm.const_array it vals in
    ignore (Llvm.define_global "GPR" v st.m)

let define_imr st =
    let it = Llvm.i32_type st.ctx in
    let fields = [| it ; it |] in
    let t = Llvm.struct_type st.ctx fields in
    let v = Llvm.const_null t in
    ignore (Llvm.define_global "IMR" v st.m)

let define_misc st =
    (* Defines miscellaneous variables *)
    let it = Llvm.i32_type st.ctx in
    let zero = Llvm.const_null it in
    ignore (Llvm.define_global "cycles" zero st.m)

external force_some : 'a option -> 'a = "%field0"

let generate_instr_block st tr_ctx f redirect instr n =
    let bb = Llvm.insert_block st.ctx n redirect in
    Llvm.position_at_end bb st.b ;
    R3000_instrs.code_for_instr tr_ctx instr ;
    bb

let bios_instruction st _ f redirect =
    let ret = Llvm.void_type st.ctx in
    let ft = Llvm.function_type ret [| |] in
    let callbios = Llvm.declare_function
        "psx_instr_callbios" ft st.m
    in
    let bb = Llvm.insert_block st.ctx "callbios" redirect in
    Llvm.position_at_end bb st.b ;
    ignore (Llvm.build_call callbios [| |] "" st.b) ;
    ignore (Llvm.build_br redirect st.b) ;
    bb

let special_instructions = [|
    bios_instruction
|]

(*
 * Object which conforms to the memory interface, passed through the
 * translation context.
 *)
let build_memory_interface st =
    let i32_type = Llvm.i32_type st.ctx in
    let mem_read_write_f size =
        let it = Llvm.integer_type st.ctx size in
        let fname typ = Printf.sprintf "psx_mem_%s_%d" typ size in
        let read_typ = Llvm.function_type it [| i32_type |] in
        let write_typ = Llvm.function_type (Llvm.void_type st.ctx)
                        [| i32_type ; it |] in
        let decl typ name =
            let fname = fname name in
            Llvm.declare_function fname typ st.m
        in
        (decl read_typ "read", (decl write_typ "write", it))
    in
    object (self)
        val funs = Array.map mem_read_write_f [| 8 ; 16 ; 32 |]

        method private int_of_size = function
        |   `Byte -> 0
        |   `Halfword -> 1
        |   `Word -> 2

        method read builder size addr =
            let f = fst funs.(self#int_of_size size) in
            let res = Llvm.build_call f [| addr |] "" builder in
            Llvm.build_zext res i32_type "" builder

        method write builder size addr value =
            let f, it = snd funs.(self#int_of_size size) in
            let value = Llvm.build_trunc value it "" builder in
            ignore (Llvm.build_call f [| addr ; value |] "" builder)
    end

let build_translation_context st redirect =
    let i32_type = Llvm.i32_type st.ctx in
    let gpr = Llvm.declare_global (Llvm.array_type i32_type 32) "GPR" st.m in
    let imr_type = Llvm.struct_type st.ctx [| i32_type ; i32_type |] in
    let imr = Llvm.declare_global imr_type "IMR" st.m in
    let jump_type = Llvm.function_type (Llvm.void_type st.ctx)
                    [| i32_type |] in
    let jump = Llvm.declare_function "psx_jump" jump_type st.m in
    let mem = build_memory_interface st in
    let pc = Llvm.declare_global i32_type "pc" st.m in {
        R3000_instrs.ctx = st.ctx ;
        b = st.b ;
        m = st.m ;
        gpr = gpr ;
        imr = imr ;
        pc = pc ;
        i32_type = i32_type ;
        jump = jump ;
        mem = mem ;
        i = -1 ;
        f = st.f ;
        endblock = redirect
    }

let generate_instrs_blocks st f redirect =
    let c = st.f.P.code in
    let si = special_instructions in
    let l = (Array.length c) + (Array.length si) in
    let tr_ctx = build_translation_context st redirect in
    let htbl = Hashtbl.create 65536 in
    Array.init l (fun i ->
        if i < Array.length si then
            si.(i) st tr_ctx f redirect
        else
            let i = i - (Array.length si) in
            let name = Printf.sprintf "at_%d" i in
            let ctx = { tr_ctx with R3000_instrs.i = i } in
            try
                let instr = D.decode c.(i) in
                if not (R3000_instrs.depends_on_pc instr) then
                    try
                        Hashtbl.find htbl instr
                    with Not_found -> (
                        let b = generate_instr_block st ctx f
                                redirect instr name in
                        Hashtbl.add htbl instr b ;
                        b
                    )
                else
                    generate_instr_block st ctx f redirect instr name
            with Invalid_argument _ ->
                redirect
    )

let generate_jumptable st f instrs =
    let instrs = Array.map (fun bb -> Llvmext.block_address f bb) instrs in
    let typ = Llvm.type_of instrs.(0) in
    let v = Llvm.const_array typ instrs in
    Llvm.define_global "jumptable" v st.m

let declare_codegate st jumptable =
    let ret = Llvm.pointer_type (Llvm.i8_type st.ctx) in
    let i32t = Llvm.i32_type st.ctx in
    let ft = Llvm.function_type ret [| i32t ; Llvm.type_of jumptable |] in
    Llvm.declare_function "psx_instr_at" ft st.m

let fill_redirect_block st f redirect instrs jumptable =
    let codegate = declare_codegate st jumptable in
    let pc = force_some (Llvm.lookup_global "pc" st.m) in
    let cycles = force_some (Llvm.lookup_global "cycles" st.m) in
    Llvm.position_at_end redirect st.b ;
    let i32t = Llvm.i32_type st.ctx in
    let c = Llvm.build_load cycles "cycles" st.b in
    let newc = Llvm.build_add c (Llvm.const_int i32t 1) "cycles" st.b in
    ignore (Llvm.build_store newc cycles st.b) ;
    let oldpc = Llvm.build_load pc "oldpc" st.b in
    let newpc = Llvm.build_add oldpc (Llvm.const_int i32t 4) "newpc" st.b in
    ignore (Llvm.build_store newpc pc st.b) ;
    ignore (Llvmext.build_indirectbr
            (Llvm.build_call codegate [| oldpc ; jumptable |] "" st.b)
            instrs
            st.b
    )

let generate_code st =
    let ft = Llvm.function_type (Llvm.void_type st.ctx) [||] in
    let f = Llvm.define_function "psx_exec_one" ft st.m in
    let endblock = Llvm.append_block st.ctx "endblock" f in
    let instrs = generate_instrs_blocks st f endblock in
    let jumptable = generate_jumptable st f instrs in
    fill_redirect_block st f (Llvm.entry_block f) instrs jumptable ;
    
    Llvm.position_at_end endblock st.b ;
    ignore (Llvm.build_ret_void st.b)

let compile st =
    define_pc st ;
    define_segs st ;
    define_gpr st ;
    define_imr st ;
    define_misc st ;
    generate_code st ;
    st
