(*
 * Translator driver.
 *
 * Copyright (c) 2009-2010 Pierre "delroth" Bourdon <delroth@gmail.com>
 *
 * This file is part of Meta-morphosis.
 *
 * Meta-morphosis is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Meta-morphosis is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Meta-morphosis.  If not, see <http://www.gnu.org/licenses/>.
 *)

module ComputerPsxFile = Psx_file.Make (Os_fs)

type cfg = {
    mutable fname : string ;
    mutable act : Psx_file.t -> out_channel -> string -> unit ;
    mutable out : string
}

let uniqid cnt =
    let id = "n" ^ (string_of_int !cnt) in
    incr cnt ;
    id

let ilist_to_s code =
    let cr = List.map R3000_decoder.to_string code in
    List.fold_left (fun a b -> a ^ "\\n" ^ b)
        (List.hd cr) (List.tl cr)

let fpf = Printf.fprintf
let rec dump_dot p cnt h cg =
    if Hashtbl.mem h cg then
        Hashtbl.find h cg
    else match !cg with
    |   Codegraph.CodeBlock (code, next) ->
            let id = uniqid cnt in
            fpf p "%s [label=\"%s\"]\n" id (ilist_to_s code) ;
            Hashtbl.add h cg id ;
            let nid = dump_dot p cnt h next in
            fpf p "%s -> %s\n" id nid ;
            id
    |   Codegraph.Branching (instrs, next) ->
            let id = uniqid cnt in
            fpf p "%s [label=\"%s\",shape=box]\n" id
                (ilist_to_s instrs) ;
            Hashtbl.add h cg id ;
            List.iter (fun n ->
                let nid = dump_dot p cnt h n in
                Hashtbl.add h n nid ;
                fpf p "%s -> %s\n" id nid
            ) next ;
            id

let dump_graph_dot f out =
    let cg = Codegraph.make f in
    let cnt = ref 0 in
    let s = Hashtbl.create 2048 in
    fpf out "digraph G {\n" ;
    ignore (dump_dot out cnt s cg) ;
    fpf out "}\n" ;
    (fun _ -> ())

let dump_backend_out typ f out =
    let module Backend = Llvmbackend in
    let outfun = match typ with
    |   `Binary -> Backend.to_binary
    |   `Bytecode -> Backend.to_bytecode
    |   `Il -> Backend.to_il
    in
    let st = Backend.init f in
    prerr_endline "Translating the executable into IL..." ;
    let o = Backend.compile st in
    outfun o out ;
    prerr_endline "Done !" ;
    match typ with
    |   `Binary -> (fun n -> Unix.chmod n 0o755)
    |   _ -> (fun _ -> ())

let parse_args () =
    let usage = "usage: translator [-dot|-bc|-il] <filename>" in
    let n = Obj.magic () in
    let cfg = { fname = n ; act = dump_backend_out `Binary ; out = n } in
    let opts = [
        "-dot", Arg.Unit (fun () -> cfg.act <- dump_graph_dot),
            "dump the code graph in dot format" ;
        "-bc", Arg.Unit (fun () -> cfg.act <- dump_backend_out `Bytecode),
            "dump the LLVM bytecode" ;
        "-il", Arg.Unit (fun () -> cfg.act <- dump_backend_out `Il),
            "dump the LLVM intermediate language" ;
        "-o", Arg.String (fun s -> cfg.out <- s),
            "path to the output file (default: ./a.out)"
    ] in
    Arg.parse opts (fun s -> cfg.fname <- s) usage ;
    if cfg.fname == n then (
        Arg.usage opts usage ;
        exit 1
    ) ;
    if cfg.out == n then (
        cfg.out <- "./a.out"
    ) ;
    cfg

let main cfg =
    let stream = Os_fs.open_ cfg.fname in
    let psxfile = ComputerPsxFile.parse stream in
    Os_fs.close stream ;
    let out = open_out cfg.out in
    let cont = cfg.act psxfile out in
    close_out out ;
    cont cfg.out

let () =
    let cfg = parse_args () in
    main cfg
