(*
 * Implements R3000 instructions in AST form in order to translate them to
 * LLVM bytecode.
 *
 * Copyright (c) 2009-2010 Pierre "delroth" Bourdon <delroth@gmail.com>
 *
 * This file is part of Meta-morphosis.
 *
 * Meta-morphosis is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Meta-morphosis is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Meta-morphosis.  If not, see <http://www.gnu.org/licenses/>.
 *)

module D = R3000_decoder
module P = Psx_file
module Reg = R3000_reg

type translation_context = {
    ctx : Llvm.llcontext ;
    b : Llvm.llbuilder ;
    m : Llvm.llmodule ;
    gpr : Llvm.llvalue ;
    imr : Llvm.llvalue ;
    pc : Llvm.llvalue ;
    i32_type : Llvm.lltype ;
    jump : Llvm.llvalue ;
    mem : Memory.t ;
    i : int ;
    f : Psx_file.t ;
    endblock : Llvm.llbasicblock
}

(* Utility functions *)
let i32 ctx v =
    Llvm.const_int ctx.i32_type v

let i32_ ctx v =
    let v = Int64.of_int32 v in
    Llvm.const_of_int64 ctx.i32_type v true

let rreg ctx r =
    if r = Reg.R0 then
        i32 ctx 0
    else
        let p = Llvm.build_bitcast ctx.gpr
                (Llvm.pointer_type ctx.i32_type) "" ctx.b in
        let p = Llvm.build_gep p [| i32 ctx (Reg.to_int r) |] "" ctx.b in
        Llvm.build_load p "" ctx.b

let wreg ctx r v =
    if r <> Reg.R0 then
        let p = Llvm.build_bitcast ctx.gpr
                (Llvm.pointer_type ctx.i32_type) "" ctx.b in
        let p = Llvm.build_gep p [| i32 ctx (Reg.to_int r) |] "" ctx.b in
        ignore (Llvm.build_store v p ctx.b)

let pc ctx =
    Llvm.build_load ctx.pc "" ctx.b

let imr_field n ctx =
    let p = Llvm.build_bitcast ctx.imr
            (Llvm.pointer_type ctx.i32_type) "" ctx.b in
    Llvm.build_gep p [| i32 ctx n |] "" ctx.b

let imr_hi = imr_field 0
let imr_lo = imr_field 1

let branch_to_end ctx =
    ignore (Llvm.build_br ctx.endblock ctx.b)

(* Addresses handling *)
let pc_of_i ctx =
    let t_seg_addr = ctx.f.P.header.P.text.P.addr in
    let i = Int32.of_int (ctx.i + 1) in
    let rel = Int32.shift_left i 2 in
    Int32.add rel t_seg_addr

let adapt_branch_address ctx addr =
    let addr = Bits.int32_of_s16 addr in
    let addr = Int32.shift_left addr 2 in
    Int32.add (pc_of_i ctx) addr

let adapt_jump_address ctx addr =
    let t_seg_addr = ctx.f.P.header.P.text.P.addr in
    let base = Int32.logand t_seg_addr 0xF0000000l in
    let addr = Int32.shift_left (Int32.of_int addr) 2 in
    Int32.add base addr

(* Generic instructions *)

let binary_reg_instr f ctx rs rt rd _ =
    let reg = rreg ctx in
    let rs, rt = (reg rs, reg rt) in
    let res = f rs rt "" ctx.b in
    wreg ctx rd res ;
    branch_to_end ctx

let binary_reg_imm_instr f ctx rs rt imm =
    let rs = rreg ctx rs in
    let res = f rs imm "" ctx.b in
    wreg ctx rt res ;
    branch_to_end ctx

let binary_reg_immi16_instr f ctx rs rt imm =
    let i16 = Llvm.const_int (Llvm.i16_type ctx.ctx) imm in
    let i32 = Llvm.build_sext i16 ctx.i32_type "" ctx.b in
    binary_reg_imm_instr f ctx rs rt i32

let binary_reg_immu16_instr f ctx rs rt imm =
    let i32 = i32 ctx imm in
    binary_reg_imm_instr f ctx rs rt i32

let generate_jump_with_val ~link ctx addr =
    begin
        match link with
        |   Some linkreg ->
                let four = i32 ctx 4 in
                let pc_plus_four = Llvm.build_add (pc ctx) four "" ctx.b in
                wreg ctx linkreg pc_plus_four
        |   None -> ()
    end ;
    ignore (Llvm.build_call ctx.jump [| addr |] "" ctx.b) ;
    branch_to_end ctx

let generate_jump ~link ctx addr =
    generate_jump_with_val ~link ctx (i32_ ctx addr)

let branch_instr link icmp ctx v1 v2 imm =
    let addr = adapt_branch_address ctx imm in
    let cmp = Llvm.build_icmp icmp v1 v2 "" ctx.b in
    let cmp_ok = Llvm.insert_block ctx.ctx "cmpok" ctx.endblock in
    ignore (Llvm.build_cond_br cmp cmp_ok ctx.endblock ctx.b) ;
    Llvm.position_at_end cmp_ok ctx.b ;
    generate_jump ~link ctx addr

let branch_two_regs_instr icmp ctx rs rt =
    branch_instr None icmp ctx (rreg ctx rs) (rreg ctx rt)

let branch_reg_zero_instr icmp ctx rs =
    let other = Llvm.const_null ctx.i32_type in
    branch_instr None icmp ctx (rreg ctx rs) other

let branch_link_reg_zero_instr icmp ctx rs =
    let other = Llvm.const_null ctx.i32_type in
    branch_instr (Some Reg.RA) icmp ctx (rreg ctx rs) other

let jump_instr ~link ctx target =
    let target = adapt_jump_address ctx target in
    let link = if link then Some Reg.RA else None in
    generate_jump ~link ctx target

let jump_reg_instr ~link ctx rs _ rd _ =
    let link = if link then Some rd else None in
    generate_jump_with_val ~link ctx (rreg ctx rs)

let div_instr ~signed ctx rs rt _ _ =
    let rt = rreg ctx rt in
    let rs = rreg ctx rs in
    let cmp = Llvm.build_icmp Llvm.Icmp.Ne rt (i32 ctx 0) "" ctx.b in
    let cmp_ok = Llvm.insert_block ctx.ctx "cmpok" ctx.endblock in
    ignore (Llvm.build_cond_br cmp cmp_ok ctx.endblock ctx.b) ;
    Llvm.position_at_end cmp_ok ctx.b ;
    let f_div, f_mod =
        if signed then
            Llvm.build_sdiv, Llvm.build_srem
        else
            Llvm.build_udiv, Llvm.build_urem
    in
    let vdiv = f_div rs rt "" ctx.b in
    let vmod = f_mod rs rt "" ctx.b in
    ignore (Llvm.build_store vdiv (imr_lo ctx) ctx.b) ;
    ignore (Llvm.build_store vmod (imr_hi ctx) ctx.b) ;
    branch_to_end ctx

let mult_instr ctx rs rt _ _ =
    let i64_type = Llvm.i64_type ctx.ctx in
    let i64 = Llvm.const_int i64_type in
    let rs = rreg ctx rs in
    let rt = rreg ctx rt in
    let rs = Llvm.build_zext rs i64_type "" ctx.b in
    let rt = Llvm.build_zext rt i64_type "" ctx.b in
    let prod = Llvm.build_mul rs rt "" ctx.b in
    let lo = Llvm.build_and prod (i32_ ctx 0xFFFFFFFFl) "" ctx.b in
    let hi = Llvm.build_lshr prod (i64 32) "" ctx.b in
    let to_i32 v = Llvm.build_trunc v ctx.i32_type "" ctx.b in
    let hi, lo = to_i32 hi, to_i32 lo in
    ignore (Llvm.build_store hi (imr_hi ctx) ctx.b) ;
    ignore (Llvm.build_store lo (imr_lo ctx) ctx.b) ;
    branch_to_end ctx

let addr_from_base_off ctx base off =
    let base = rreg ctx base in
    let off = i32_ ctx (Bits.int32_of_s16 off) in
    Llvm.build_add base off "" ctx.b

let mem_load_instr ~signed ~size ctx rs rt imm =
    let addr = addr_from_base_off ctx rs imm in
    let read = ctx.mem#read ctx.b size addr in
    let read =
        if signed then
            Llvm.build_sext read ctx.i32_type "" ctx.b
        else
            Llvm.build_zext read ctx.i32_type "" ctx.b
    in
    wreg ctx rt read ;
    branch_to_end ctx

let mem_store_instr ~size ctx rs rt imm =
    let addr = addr_from_base_off ctx rs imm in
    ctx.mem#write ctx.b size addr (rreg ctx rt) ;
    branch_to_end ctx

let store_from_imr_field field ctx _ _ rd _ =
    let f = match field with `Hi -> imr_hi | `Lo -> imr_lo in
    wreg ctx rd (Llvm.build_load (f ctx) "" ctx.b) ;
    branch_to_end ctx

let store_to_imr_field field ctx rs _ _ _ =
    let f = match field with `Hi -> imr_hi | `Lo -> imr_lo in
    ignore (Llvm.build_store (rreg ctx rs) (f ctx) ctx.b) ;
    branch_to_end ctx

let build_nor v1 v2 _ builder =
    let or_ = Llvm.build_or v1 v2 "" builder in
    Llvm.build_not or_ "" builder

let shift_imm_instr f ctx _ rt rd shift =
    wreg ctx rd (f (rreg ctx rt) (i32 ctx shift) "" ctx.b) ;
    branch_to_end ctx

let shift_reg_instr f ctx rs rt rd _ =
    wreg ctx rd (f (rreg ctx rt) (rreg ctx rs) "" ctx.b) ;
    branch_to_end ctx

let store_cmp_result icmp ctx out v1 v2 =
    let v = Llvm.build_icmp icmp v1 v2 "" ctx.b in
    let v = Llvm.build_zext v ctx.i32_type "" ctx.b in
    wreg ctx out v ;
    branch_to_end ctx

let store_reg_cmp_result icmp ctx rs rt rd _ =
    store_cmp_result icmp ctx rd (rreg ctx rs) (rreg ctx rt)

let store_imm_cmp_result icmp ctx rs rt imm =
    store_cmp_result icmp ctx rt (rreg ctx rs) (i32 ctx imm)

let array_at ctx arr i =
    let llvmarr = Llvm.const_array ctx.i32_type arr in
    let p = Llvm.define_global "" llvmarr ctx.m in
    Llvm.set_global_constant true p ;
    Llvm.set_visibility Llvm.Visibility.Hidden p ;
    let p = Llvm.build_gep p [| i32 ctx 0 ; i |] "" ctx.b in
    Llvm.build_load p "" ctx.b

let load_word_left_right arrmask arrshift sf ctx rs rt imm =
    let addr = addr_from_base_off ctx rs imm in
    let shift = Llvm.build_and addr (i32 ctx 3) "" ctx.b in
    let not_3 = Llvm.build_not (i32 ctx 3) "" ctx.b in
    let real_addr = Llvm.build_and addr not_3 "" ctx.b in
    let mem = ctx.mem#read ctx.b `Word real_addr in
    let mask = array_at ctx (Array.map (i32_ ctx) arrmask) shift in
    let shift = array_at ctx (Array.map (i32 ctx) arrshift) shift in
    let left = Llvm.build_and (rreg ctx rt) mask "" ctx.b in
    let right = sf mem shift "" ctx.b in
    let v = Llvm.build_or left right "" ctx.b in
    wreg ctx rt v ;
    branch_to_end ctx

let store_word_left_right arrmask arrshift sf ctx rs rt imm =
    let addr = addr_from_base_off ctx rs imm in
    let shift = Llvm.build_and addr (i32 ctx 3) "" ctx.b in
    let not_3 = Llvm.build_not (i32 ctx 3) "" ctx.b in
    let real_addr = Llvm.build_and addr not_3 "" ctx.b in
    let mem = ctx.mem#read ctx.b `Word real_addr in
    let mask = array_at ctx (Array.map (i32_ ctx) arrmask) shift in
    let shift = array_at ctx (Array.map (i32 ctx) arrshift) shift in
    let left = Llvm.build_and mem mask "" ctx.b in
    let right = sf (rreg ctx rt) shift "" ctx.b in
    let v = Llvm.build_or left right "" ctx.b in
    ctx.mem#write ctx.b `Word real_addr v ;
    branch_to_end ctx

(* Instructions *)

let handle_Add = binary_reg_instr Llvm.build_add
let handle_Addi = binary_reg_immi16_instr Llvm.build_add
let handle_Addiu = binary_reg_immi16_instr Llvm.build_add
let handle_Addu = binary_reg_instr Llvm.build_add

let handle_And = binary_reg_instr Llvm.build_and
let handle_Andi = binary_reg_immu16_instr Llvm.build_and

let handle_Beq = branch_two_regs_instr Llvm.Icmp.Eq
let handle_Bgez = branch_reg_zero_instr Llvm.Icmp.Sge
let handle_Bgezal = branch_link_reg_zero_instr Llvm.Icmp.Sge
let handle_Bgtz = branch_reg_zero_instr Llvm.Icmp.Sgt
let handle_Blez = branch_reg_zero_instr Llvm.Icmp.Sle
let handle_Bltz = branch_reg_zero_instr Llvm.Icmp.Slt
let handle_Bltzal = branch_link_reg_zero_instr Llvm.Icmp.Slt
let handle_Bne = branch_two_regs_instr Llvm.Icmp.Ne

let handle_Break ctx _ _ = branch_to_end ctx

let handle_Div = div_instr ~signed:true
let handle_Divu = div_instr ~signed:false

let handle_J = jump_instr ~link:false
let handle_Jal = jump_instr ~link:true
let handle_Jr = jump_reg_instr ~link:false
let handle_Jalr = jump_reg_instr ~link:true

let handle_Lb = mem_load_instr ~signed:true ~size:`Byte
let handle_Lbu = mem_load_instr ~signed:false ~size:`Byte
let handle_Lh = mem_load_instr ~signed:true ~size:`Halfword
let handle_Lhu = mem_load_instr ~signed:false ~size:`Halfword
let handle_Lw = mem_load_instr ~signed:false ~size:`Word

let handle_Lui ctx _ rt imm =
    wreg ctx rt (i32_ ctx (Int32.shift_left (Int32.of_int imm) 16)) ;
    branch_to_end ctx

let handle_Mfhi = store_from_imr_field `Hi
let handle_Mflo = store_from_imr_field `Lo
let handle_Mthi = store_to_imr_field `Hi
let handle_Mtlo = store_to_imr_field `Lo

let handle_Mult = mult_instr
let handle_Multu = mult_instr

let handle_Nor = binary_reg_instr build_nor
let handle_Or = binary_reg_instr Llvm.build_or
let handle_Ori = binary_reg_immu16_instr Llvm.build_or

let handle_Sb = mem_store_instr ~size:`Byte
let handle_Sh = mem_store_instr ~size:`Halfword
let handle_Sw = mem_store_instr ~size:`Word

let handle_Sub = binary_reg_instr Llvm.build_sub
let handle_Subu = binary_reg_instr Llvm.build_sub

let handle_Syscall ctx _ _ = branch_to_end ctx

let handle_Sll = shift_imm_instr Llvm.build_shl
let handle_Sra = shift_imm_instr Llvm.build_ashr
let handle_Srl = shift_imm_instr Llvm.build_lshr

let handle_Sllv = shift_reg_instr Llvm.build_shl
let handle_Srav = shift_reg_instr Llvm.build_ashr
let handle_Srlv = shift_reg_instr Llvm.build_lshr

let handle_Slt = store_reg_cmp_result Llvm.Icmp.Slt
let handle_Sltu = store_reg_cmp_result Llvm.Icmp.Ult
let handle_Slti = store_imm_cmp_result Llvm.Icmp.Slt
let handle_Sltiu = store_imm_cmp_result Llvm.Icmp.Ult

let handle_Xor = binary_reg_instr Llvm.build_xor
let handle_Xori = binary_reg_immu16_instr Llvm.build_xor

let handle_Lwl = load_word_left_right
    [| 0xFFFFFFl ; 0xFFFFl ; 0xFFl ; 0l |]
    [| 24 ; 16 ; 8 ; 0 |]
    Llvm.build_shl

let handle_Lwr = load_word_left_right
    [| 0l ; 0xFF000000l ; 0xFFFF0000l ; 0xFFFFFF00l |]
    [| 0 ; 8 ; 16 ; 24 |]
    Llvm.build_lshr

let handle_Swl = store_word_left_right
    [| 0xFFFFFF00l ; 0xFFFF0000l ; 0xFF000000l ; 0l |]
    [| 24 ; 16 ; 8 ; 0 |]
    Llvm.build_lshr

let handle_Swr = store_word_left_right
    [| 0l ; 0xFFl ; 0xFFFFl ; 0xFFFFFFl |]
    [| 0 ; 8 ; 16 ; 24 |]
    Llvm.build_shl

module I = struct
    open D.I

    let call ctx f p =
        f ctx p.rs p.rt p.imm
end

module J = struct
    open D.J

    let call ctx f p =
        f ctx p.target
end

module R = struct
    open D.R

    let call ctx f p =
        f ctx p.rs p.rt p.rd p.shift
end

module S = struct
    open D.S

    let call ctx f p =
        f ctx p.code p.funct
end

module B = struct
    open D.B

    let call ctx f p =
        f ctx p.rs p.imm
end

let code_for_instr ctx = function
#define R3000_INSTR(name, args) \
|   D.name p -> args.call ctx handle_##name p
#include "../disasm/instrs.inl"
#undef R3000_INSTR

module StringSet = Set.Make (String)

let instrs_depending_on_pc =
    let l = [
        "Beq" ; "Bgez" ; "Bgezal" ; "Bgtz" ;
        "Blez" ; "Bltz" ; "Bltzal" ; "Bne"
    ] in
    let empty = StringSet.empty in
    List.fold_left (fun set x -> StringSet.add x set) empty l

let instr_name = function
#define R3000_INSTR(name, params) \
|   D.name _ -> #name
#include "../disasm/instrs.inl"
#undef R3000_INSTR

let depends_on_pc i =
    let name = instr_name i in
    StringSet.mem name instrs_depending_on_pc
